/*
 ***************************************************************************
 *        Copyright 2002 Compaq Information Technologies Group, L.P.       *
 ***************************************************************************
 *
 *  Title:    OS Utilities
 *
 *  Module:   cevtw_uw700.c
 *
 *  Version:  1.0
 *
 *  Date:     11/30/2001
 *
 *  Author:   Patrick M. Schoeller
 *
 ***************************************************************************
 * Description:
 * 
 * The purpose of this file is to isolate OS specific details from the
 * primary code path, thus enabling an 80% common code base.  Examples of
 * functions to include in here are Physical Memory Mapping, Kernel Memory
 * allocation, device discovery, OS specific entry points, etc.
 *
 ***************************************************************************/

#include "oswrap.h"

/*
** We need "sprintf" defined for all OS's except Linux.
*/
#ifdef sprintf
#undef sprintf
#endif

#include "cpqevtw_linux.h"
#include "cpqevtq.h"
#include "cpqevtc.h"


typedef int(* PRIB_ENTRY)(VOID *pCridData);

extern PCEVTDATA gpInstanceData;

ULONG cpqwEventHandler( LONG lAction, LONG lMatchCode,
                        LONG lEvtNum, VOID *pEvt );

ULONG cevtw_reg_iml_handler(ULONG ulAction, PIML_LOGGER pImlHndlr);

int (*callCridLog)(void* Log) = NULL;
int (*cevtw_CridLogEntry)(void* Log) = NULL;

EXPORT_SYMBOL_NOVERS(cpqwEventHandler);
EXPORT_SYMBOL_NOVERS(callCridLog);
EXPORT_SYMBOL_NOVERS(cevtw_CridLogEntry);
EXPORT_SYMBOL_NOVERS(cevtw_reg_iml_handler);


static int cevt_major = 0;

static struct file_operations cevt_ops = {	
	owner:	THIS_MODULE,
};						

/*
** FUNCTION DEFINITIONS
*/

static int cevt_init_one(void);
static int cevt_remove_one(void);
static void cevtw_free_instance(CEVTDATA *pInstance);
void cevtw_lock_iml_queue(PCEVTDATA pInstance);
void cevtw_unlock_iml_queue(PCEVTDATA pInstance);
static void cevtw_free_instance(PCEVTDATA pInstance);
static int cevtw_RIB_interface(VOID *pCridData);




/*********************************************************************

@func  init_module

Routine Description:
This is a Linux specific routine which is called when the module
is loaded by the Linux operating system.


Return Value:


***********************************************************************/
int init_module(void)
{
	cevt_init_one();

	cevt_major = register_chrdev(0, "cpqhealth", &cevt_ops);
   if (cevt_major < 0) {
      printk(KERN_WARNING "cevt:  Not able to get Major Number.  Error (%d)\n",
             cevt_major);
      return (cevt_major);
   }  /* endif */

   inter_module_register("cevtw_reg_iml_handler", THIS_MODULE, cevtw_reg_iml_handler);

   return (0);

}  /* end init_module */


/*********************************************************************

@func  cleanup_module

Routine Description:
cleanup_module() entry point is called prior to unloading.
This will not be called if any channels are open.


Return Value:


***********************************************************************/
void cleanup_module(void)
{
	cevt_remove_one();
	unregister_chrdev(cevt_major, "cpqhealth");
   inter_module_unregister("cevtw_reg_iml_handler");

}   /* end cleanup_module */


/*********************************************************************

@func  cevt_init_one

Routine Description:
This routine will initalize the device and the device driver.


Return Value:


***********************************************************************/
static int cevt_init_one(void)
{
   int            rv;
   struct pci_dev *pPciDev = NULL;
   PCEVTDATA      pInstance = NULL;
   POSDATA        pOsData = NULL;


/* Allocate and initialize a new instance structure. */
   gpInstanceData = (PCEVTDATA)kmalloc(sizeof (CEVTDATA), GFP_KERNEL);

   if (!gpInstanceData) return(-EAGAIN);

   pInstance = gpInstanceData;     /* Set local pointer for convenience */
   memset(pInstance, 0, sizeof(CEVTDATA));

   pInstance->pOsData = (PVOID)kmalloc(sizeof (OSDATA), GFP_KERNEL);
   if (!pInstance->pOsData) return(-EAGAIN);

   pOsData = (POSDATA)pInstance->pOsData;
   memset(pOsData, 0, sizeof(OSDATA));

   spin_lock_init(&pOsData->imlq_lock);
   spin_lock_init(&pOsData->ribq_lock);

   cevtw_CridLogEntry = (PRIB_ENTRY)cevtw_RIB_interface;

   if ( (cevtc_common_init(pInstance)) ) {
      cevtw_free_instance(pInstance); 
      return (EOPNOTSUPP);
   }  /* endif */


/* Announce our arrival! */
   printk("cevt:   Compaq Event Logging Driver (rev %s)\n", EFS_VER);

   return (SUCCESS);

}  /* cevt_init_one */


/*********************************************************************

@func  cevt_remove_one

Routine Description:
This routine will tear down the driver when the driver is unloaded.


Return Value:


***********************************************************************/
static int cevt_remove_one(void)
{
#ifdef THIS_NEEDS_TO_BE_DONE
	cevtc_common_shutdown(gpInstanceData);
#endif

   cevtw_CridLogEntry = NULL;
	cevtw_free_instance(gpInstanceData);
	return (0);
}   /* end cevt_remove_one */




/*********************************************************************
** 
** @func  cevtw_free_instance
** 
** Routine Description:
** This routine will clean up all the various things which were started
** or allocated during the initialization phase of this driver.
** 
** Return Value:
** Standard DDI Return codes
** 
***********************************************************************/
static void cevtw_free_instance(PCEVTDATA pInstance)
{
   int32_t     ulMemSize;
   int32_t     i;

/*
** Let's make sure we haven't already done this once already.
*/
   if (pInstance == NULL)
      return;

   if (pInstance->pOsData)
      kfree(pInstance->pOsData);
   kfree(pInstance);
   pInstance = NULL;
   gpInstanceData = NULL;


} /* END cevtw_free_instance */




/***************************************************************************
 ****************       START OF "MUST HAVE" FUNCTION    *******************
 ***************************************************************************/


/*********************************************************************
** 
** @func  cevtw_RIB_interface
** 
** Routine Description:
** This routine is used as a "shim" between the "casm" driver and the
** RIB ("crid") driver.  
** 
** Return Value:
** 
***********************************************************************/
int cevtw_RIB_interface(VOID *pCridData)
{
   int rv;

   gpInstanceData->pRIBLogger = (PRIB_LOGGER)callCridLog;

   if (gpInstanceData->pRIBLogger == NULL) 
      rv = cevtw_find_RIB_entry(gpInstanceData);

   if (gpInstanceData->pRIBLogger)
       gpInstanceData->pRIBLogger(pCridData);

   return(SUCCESS);

}  /* end cevtw_RIB_interface */


/*********************************************************************
** 
** @func  cevtw_reg_iml_handler
** 
** Routine Description:
** This routine is used to register the "cevt" device driver's IML
** logging routine (handler).  This is the shim used to support the
** legacy storage drivers.
** 
** Return Value:
**       0   : SUCCESS
**       1   : FAILURE
** 
***********************************************************************/
ULONG cevtw_reg_iml_handler(ULONG ulAction, PIML_LOGGER pImlHndlr)
{
   ULONG  retVal = SUCCESS;

   cevtw_lock_iml_queue(gpInstanceData);

   if (gpInstanceData->pImlLogger) {
      retVal = FAILURE;

   } else if (ulAction == CEVT_ADD_HANDLER) {
      gpInstanceData->pImlLogger = pImlHndlr;
      cevtc_common_iml_reg(gpInstanceData);

   } else if (ulAction == CEVT_REMOVE_HANDLER) {
      gpInstanceData->pImlLogger = NULL;
      gpInstanceData->pRIBLogger = NULL;
   /*
   ** We setup up for the HEALTH RELOAD state in the
   ** event the health driver is stopped and started.
   */
      gpInstanceData->ulCfgState |= CEVT_HEALTH_RELOAD;

   } else {
      retVal = FAILURE;
   }  /* endif */

   cevtw_unlock_iml_queue(gpInstanceData);

   return (retVal);

}  /* end cevtw_reg_iml_handler */


/*********************************************************************
** 
** @func  cpqwEventHandler
** 
** Routine Description:
** This routine is unusual in that we have to have it because
** DDI 8 device driver use a "#pragma weak" to locate the 
** health driver for event logging.  This is here purely to
** support legacy device drivers.
** 
** Return Value:
**       0   : SUCCESS
**       1   : FAILURE
** 
***********************************************************************/
ULONG cpqwEventHandler( LONG lAction, LONG lMatchCode,
                              LONG lEvtNum, VOID *pEvt )
{

   return ( (ULONG)cevtc_log_iml_event(lAction, lMatchCode, lEvtNum, pEvt) );

}  /* end cpqwEventHandler */

/*********************************************************************
** 
** @func  cevtw_malloc
** 
** Routine Description:
** This routine will allocate kernel memory for 
** driver use.  This is NOT mapped in memory but
** allocated from the kernel pools.
** 
** Return Value:
** Pointer to allocated memory or NULL if this fails.
** 
***********************************************************************/
PVOID cevtw_malloc(ULONG ulSize)
{
   PVOID  pVMem;
   ULONG  ulOrder;

   printk(KERN_DEBUG "cevtw_malloc:  Asking for %d bytes\n", ulSize);
   ulOrder = get_order(ulSize);
   if (ulOrder > 5) {
      pVMem = (PVOID)__get_free_pages(GFP_ATOMIC, ulOrder);
   } else {
      pVMem = (PVOID)kmalloc(ulSize, GFP_ATOMIC);
   }  /* endif */

   printk(KERN_DEBUG "cevtw_malloc:  %p (size %d)\n",pVMem, ulSize);
   return (pVMem);

}  /* end cevtw_malloc */


/*********************************************************************
** 
** @func  cevtw_free
** 
** Routine Description:
** This routine will free kernel memory previously
** allocated for driver use.  This is NOT mapped in memory but
** allocated from the kernel pools.
** 
** Return Value:
** Nothing needed to return
** 
***********************************************************************/
VOID cevtw_free(PVOID pVAddr, ULONG ulSize)
{
   ULONG  ulOrder;

   printk(KERN_DEBUG "cevtw_free:  %p (size %d)\n",pVAddr, ulSize);

   ulOrder = get_order(ulSize);
   if (ulOrder > 5) {
      free_pages((ULONG)pVAddr, ulOrder);
   } else {
      kfree (pVAddr);
   }  /* endif */

}  /* end cevtw_free */

/*********************************************************************
** 
** @func  cevtw_memcpy
** 
** Routine Description:
** This is an OS specific routine which will copy from Source to 
** destination for a certain amount of bytes.  No Alignment is
** assumed so individual OS's may need to double buffer to assure
** alignment if required.
** 
** Return Value:
** 
** Nothing
** 
***********************************************************************/
ULONG cevtw_memcpy(char *dest, char *source, int length)
{
        memcpy(dest, source,  length);
        return ( (ULONG) dest );

}  /* end cevtw_memcpy */



/*********************************************************************
** 
** @func  cevtw_strcpy
** 
** Routine Description:
** This is an OS specific routine which will copy from Source to 
** destination until a NULL character is received.  No alignment is
** assumed so individual OS's may need to double buffer to assure
** alignment if required.
** 
** Return Value:
** Byte pointer to the Destination
** 
***********************************************************************/
char *cevtw_strcpy (char *s1, char *s2)
{

   return ( (char *) strcpy(s1,s2));

} /* end cevtw_strcpy */



/*********************************************************************
** 
** @func  cevtw_strcat
** 
** Routine Description:
** This is an OS specific routine which will concatenate from Source to 
** destination until a NULL character is received.  No alignment is
** assumed so individual OS's may need to double buffer to assure
** alignment if required.
** 
** Return Value:
** 
** Byte pointer to the destination.
** 
***********************************************************************/
char *cevtw_strcat (char *s1, char *s2)
{

   return ( (char *) strcat(s1,s2));

}  /* end cevtw_strcat */


/*********************************************************************
** 
** @func  cevtw_strlen
** 
** Routine Description:
** This is an OS specific routine which return the length of a null
** terminated string. No alignment is assumed so individual OS's may 
** need to double buffer to assure alignment if required.
** 
** Return Value:
** 
** String Length
** 
***********************************************************************/
ulong cevtw_strlen (char *s1)
{

   return ( (ulong) strlen(s1) );

} /* end cevtw_strlen */



/*********************************************************************
** 
** @func  cevtw_memzero
** 
** Routine Description:
** This OS Specific routine will "zero out" a memory location for the
** length specified.
** No alignment is assumed so individual OS's may need to double buffer 
** to assure alignment if required.
** 
** Return Value:
** 
** VOID pointer to memory addressed passed in.
** 
***********************************************************************/
void *cevtw_memzero(void *s, int n)
{
   memset(s,0,n);
   return (s);
}  /* end cevtw_memzero */


/*********************************************************************
** 
** @func  cevtw_memset
** 
** Routine Description:
** This is an OS Specific routine which will set memory to a particular
** value as passed in. We will use cevtw_memzero if the value is zero
** to gain the efficency of the SCO UnixWare library routines.
** No alignment is assumed so individual OS's may need to double buffer 
** to assure alignment if required.
** 
** Return Value:
** VOID pointer to memory addressed passed in.
**
**
***********************************************************************/
void * cevtw_memset(  void *s, register int c, register int n)
{
   memset(s,c,n);
   return (s);

}  /* end cevtw_memset */

/*********************************************************************
** 
** @func  cevtw_strncmp
** 
** Routine Description:
** This is an OS Specific routine which will compare two strings for
** a specified length.  Follows the standard C Library for "strncmp".
** No alignment is assumed so individual OS's may need to double buffer 
** to assure alignment if required.
** 
** Return Value:
** VOID pointer to memory addressed passed in.
** 
***********************************************************************/
int cevtw_strncmp(register char *s1, register char *s2, register int length)
{
   char *end;

   for (end = s1 + length - 1; s1 < end && *s1 - *s2 == 0; s1++, s2++);
   return((int)(*s1 - *s2));

} /* end cevtw_strncmp */


/*********************************************************************
** 
** @func  cevtw_printf
** 
** Routine Description:
** This is an OS Specific routine which emulate a printf command for a
** kernel driver.  In the case of UnixWare, this is the cmn_err routine.
** The arguments to this driver are the following:
** 
** ulErrorLevel   :  The error level which to log.  
** pString        :  The pointer to the string to pass to "cmn_err".
** ppParms        :  The Address of where the parameters are on the stack.
**                   This is set by the calling routine since the parameters
**                   are on the stack before the following the "pString"
**                   pointer.
** pInstance      :  The "Instance Data" for this driver which is passed to
**                   all routines in this module.
** 
** 
** 
** Return Value:
** Nothing 
** 
***********************************************************************/
VOID cevtw_printf( ULONG ulErrorLevel, PCHAR pString, 
                   PULONG ppParms, PCEVTDATA pInstance )
{
   PULONG p1;
   PULONG p2;
   PULONG p3;
   PULONG p4;
   PULONG p5;
   PULONG p6;
   PULONG p7;
   PULONG p8;
   PULONG p;

   ULONG ulStringLen;
   PCHAR pAlloc;
   PCHAR pCmnMsg;

   /*
   ** Insert the string "cevt: " to the OS_OutErr() messages to 
   ** identify the driver.
   */
   ulStringLen = (cevtw_strlen(pString)) + 21;
   pAlloc = (PCHAR)cevtw_malloc(ulStringLen);
   if (pAlloc)
   {
      pCmnMsg = pAlloc;

   /*
   ** Now make the call to printk with the appropriate error level set
   */

      switch (ulErrorLevel) {

         case CEVT_KERN_INFO:
            cevtw_strcpy(pCmnMsg, KERN_NOTICE);
            cevtw_strcat(pCmnMsg, " NOTICE: cevt: ");
            break;

         case CEVT_KERN_WARN:
            cevtw_strcpy(pCmnMsg, KERN_WARNING);
            cevtw_strcat(pCmnMsg, " WARNING: cevt: ");
            break;

         case CEVT_KERN_DEBUG:
            cevtw_strcpy(pCmnMsg, KERN_DEBUG);
            cevtw_strcat(pCmnMsg, " DEBUG: cevt: ");
            break;

         case CEVT_KERN_CRITICAL:
            cevtw_strcpy(pCmnMsg, KERN_CRIT);
            cevtw_strcat(pCmnMsg, " CRITICAL: cevt: ");
            break;

         case CEVT_KERN_PANIC:
            cevtw_strcpy(pCmnMsg, KERN_EMERG);
            cevtw_strcat(pCmnMsg, " PANIC: cevt: ");
            break;

         default:
            printk(KERN_WARNING "ERROR! cevt:  Invalid Error Level from Common Code:  %d\n",
                                ulErrorLevel );
            cevtw_strcpy(pCmnMsg, KERN_WARNING);
            cevtw_strcat(pCmnMsg, " ERROR: cevt: ");
            break;

      }  /* endswitch */

      cevtw_strcat(pCmnMsg, pString);
      cevtw_strcat(pCmnMsg, "\n");
   } else {
      pCmnMsg = pString;
   }  /* endif */

/*
** Fixup the pointers so they point to upto 8 32bit input parameters.
** The way cmn_err works, it expects the parameters to be located on the
** stack immediately following the pointer.  We will need to copy the
** parameters from the original location on the stack (as pointed to
** by ppParms) to the current stack pointer location when we make the
** call to cmn_err.
*/

   p = p1 = ppParms;
   p2 = ++p;
   p3 = ++p;
   p4 = ++p;
   p5 = ++p;
   p6 = ++p;
   p7 = ++p;
   p8 = ++p;

/*
** Now make the call to printk
*/

   printk(pCmnMsg,*p1,*p2,*p3,*p4,*p5,*p6,*p7,*p8); 

   if (pAlloc)
      cevtw_free(pAlloc, ulStringLen);

}  /* end cevtw_printf */


/*********************************************************************
** 
** @func  cevtw_lock_iml_queue
** 
** Routine Description:
** This routine is OS specific and will take a synchronization lock for
** access to the IML queue.
** 
** Return Value:
** Nothing needed to return
** 
***********************************************************************/
void cevtw_lock_iml_queue(PCEVTDATA pInstance)
{
    POSDATA pOsData = (POSDATA)pInstance->pOsData;
    spin_lock_irqsave(&pOsData->imlq_lock, pOsData->imlq_pl);
} /* cevtw_lock_iml_queue */


/*********************************************************************
** 
** @func  cevtw_unlock_iml_queue
** 
** Routine Description:
** This routine is OS specific and will free a synchronization lock for
** the IML queue.
** 
** Return Value:
** Nothing needed to return
** 
***********************************************************************/
void cevtw_unlock_iml_queue(PCEVTDATA pInstance)
{
    POSDATA pOsData = (POSDATA)pInstance->pOsData;

    spin_unlock_irqrestore(&pOsData->imlq_lock, pOsData->imlq_pl);
} /* cevtw_unlock_iml_queue */


/*********************************************************************
** 
** @func  cevtw_lock_rib_queue
** 
** Routine Description:
** This routine is OS specific and will take a synchronization lock for
** the RIB (RILOE) queue.
** 
** Return Value:
** Nothing needed to return
** 
***********************************************************************/
void cevtw_lock_rib_queue(PCEVTDATA pInstance)
{
    POSDATA pOsData = (POSDATA)pInstance->pOsData;

    spin_lock_irqsave(&pOsData->ribq_lock, pOsData->ribq_pl);
} /* cevtw_lock_rib_queue */


/*********************************************************************
** 
** @func  cevtw_unlock_rib_queue
** 
** Routine Description:
** This routine is OS specific and will free a synchronization lock for
** the RIB (RILOE) queue.
** 
** Return Value:
** Nothing needed to return
** 
***********************************************************************/
void cevtw_unlock_rib_queue(PCEVTDATA pInstance)
{
    POSDATA pOsData = (POSDATA)pInstance->pOsData;

    spin_unlock_irqrestore(&pOsData->ribq_lock, pOsData->ribq_pl);

} /* cevtw_unlock_rib_queue */



/*********************************************************************
** 
** @func  cevtw_update_entry_points
** 
** Routine Description:
** This routine will register the entry points for other drivers to
** log events to the Compaq Integrated Management Log (IML).
** Initializes the dynamic event log support advertisement
** mechanism. This is used by other drivers who wish to
** call the event log handler but do not wish to declare
** it as an external symbol.
** 
** Return Value:
**       0   : SUCCESS
**       1   : FAILURE
** 
***********************************************************************/
int cevtw_update_entry_points(PCEVTDATA pInstance)
{

   return (SUCCESS);

}  /* end cevtw_update_entry_points */



/*********************************************************************
** 
** @func  cevtw_find_RIB_entry
** 
** Routine Description:
** This routine will hook the legacy "crid" entry points used to
** log event to the Compaq Remote Insight Board.
** 
** Return Value:
**       0   : SUCCESS
**       1   : FAILURE
** 
***********************************************************************/
int cevtw_find_RIB_entry(PCEVTDATA pInstance)
{

   int           retVal = FAILURE;

   if (callCridLog != NULL) {
      gpInstanceData->pRIBLogger = (PRIB_LOGGER)callCridLog;
      retVal = SUCCESS;

   }  /* endif */

   return (retVal);

}  /* end cevtw_find_RIB_entry */


